import typing
from abc import ABC, abstractmethod


class AbstractCaster(ABC):
    @abstractmethod
    def cast(self, value: str) -> typing.Any:
        pass


class IntCaster(AbstractCaster):
    def cast(self, value: str) -> int:
        return int(value)

class FloatCaster(AbstractCaster):
    def cast(self, value: str) -> float:
        return float(value)

class BoolCaster(AbstractCaster):
    def __init__(self):
        self.true = {"true", "yes", "on", "enable", "1"}
        self.false = {"false", "no", "off", "disable", "0"}

    def cast(self, value: str) -> bool:
        if value in self.true:
            return True
        elif value in self.false:
            return False

        raise ValueError("value provided cannot be casted to bool")

class ListCaster(AbstractCaster):
    def __init__(self, sep: str = ",", item_caster: AbstractCaster | None = None):
        self.sep = sep
        self.item_caster = item_caster

    def cast(self, value: str) -> list[typing.Any]:
        str_items = value.split(self.sep)

        if self.item_caster:
            casted_items: list[typing.Any] = []
            for item in str_items:
                try:
                    casted = self.item_caster.cast(item)
                    casted_items.append(casted)
                except Exception as e:
                    raise ValueError(f'failed to cast list item "{item}" using {self.item_caster.__class__.__name__}') from e
            return casted_items

        return str_items


to_int = IntCaster()
to_float = FloatCaster()
to_bool = BoolCaster()
to_list = ListCaster()

# todo: write more basic casters



from .caster import AbstractCaster
from .provider import AbstractProvider


class CastingError(Exception):
    field_name: str
    caster: AbstractCaster
    exception: Exception

    def __init__(self, field_name: str, raw_value: str, caster: AbstractCaster, exception: Exception):
        self.field_name = field_name
        self.raw_value = raw_value
        self.caster = caster
        self.exception = exception

        super().__init__(
            f'exception occurred when casting {field_name} value "{raw_value}" '
            f"using {caster.__class__.__name__}: {exception}"
        )


class FieldValueNotProvidedError(Exception):
    def __init__(self, field_name: str, provider: AbstractProvider):
        self.field_name = field_name
        self.provider = provider

        super().__init__(f"{field_name} was not provided by {provider.__class__.__name__}, but was expected")


class FieldConflictError(Exception):
    field_name: str
    file_field_name: str

    def __init__(self, field_name: str, file_field_name: str, provider: AbstractProvider):
        self.field_name = field_name
        self.file_field_name = file_field_name

        super().__init__(
            f"values for both {field_name} and {file_field_name} "
            f"fields were provided by {provider.__class__.__name__}, "
            "expected only value for one of them"
        )



import unittest
from unittest.mock import MagicMock, patch, mock_open
from minicfg import Minicfg, minicfg_prefix, minicfg_provider
from minicfg.field import Field, AbstractField, _NOT_SET
from minicfg.provider import AbstractProvider, EnvProvider
from minicfg.exceptions import FieldConflictError, FieldValueNotProvidedError


# Mock Field and AbstractProvider for testing purposes
class MockField(Field):
    def __init__(self, name=None, default=_NOT_SET, caster=None, attach_file_field=False):
        super().__init__(name, default, caster, attach_file_field)

    def name_with_prefix(self, prefix):
        return f"{prefix}{self.name}" if prefix else self.name


class MockProvider(AbstractProvider):
    def __init__(self, values):
        self.values = values

    def get(self, key: str) -> str | None:
        return self.values.get(key)


class TestMinicfg(unittest.TestCase):

    def setUp(self):
        self.provider = MockProvider({
            "testfield1": "value1",
            "testfield2": "value2",
            "testfield1_FILE": "file_content"
        })
        self.minicfg = Minicfg(prefix="test", provider=self.provider)

    @patch('minicfg.minicfg._read_raw_value_from_file', return_value="file_content")
    def test_populate(self, mock_read):
        field = MockField(name="field1")
        self.minicfg.field1 = field

        self.minicfg.populate()

        self.assertTrue(self.minicfg.is_populated)
        self.assertEqual("value1", self.minicfg.field1.value)

    def test_field_conflict_error(self):
        field = MockField(name="field1", attach_file_field=True)
        self.minicfg.field1 = field
        file_field = MockField(name="field1_FILE")
        self.minicfg.field1_FILE = file_field

        with self.assertRaises(FieldConflictError):
            self.minicfg.populate()

    def test_field_value_not_provided_error(self):
        field = MockField(name="field3")
        self.minicfg.field3 = field

        with self.assertRaises(FieldValueNotProvidedError):
            self.minicfg.populate()

    def test_dict_method(self):
        field = MockField(name="field1")
        self.minicfg.field1 = field
        self.minicfg.populate()

        result = self.minicfg.dict()
        self.assertIn("field1", result)
        self.assertEqual(result["field1"], "value1")

    def test_dict_not_populated(self):
        with self.assertRaises(RuntimeError):
            self.minicfg.dict()

    def test_minicfg_prefix_decorator(self):
        @minicfg_prefix("prefix")
        class PrefixedMinicfg(Minicfg):
            pass

        obj = PrefixedMinicfg()
        self.assertEqual(obj._prefix, "prefix_")

    def test_minicfg_provider_decorator(self):
        mock_provider = MagicMock()

        @minicfg_provider(mock_provider)
        class ProviderMinicfg(Minicfg):
            pass

        obj = ProviderMinicfg()
        self.assertEqual(obj._default_provider, mock_provider)


if __name__ == "__main__":
    unittest.main()



import os
import unittest
from unittest.mock import patch

from minicfg.provider import EnvProvider


class TestEnvProvider(unittest.TestCase):

    def setUp(self):
        # Save the current state of environment variables
        self.provider = EnvProvider()
        self.original_environ = dict(os.environ)

    def tearDown(self):
        # Restore the original environment variables
        os.environ.clear()
        os.environ.update(self.original_environ)

    def test_get_existing_env_var(self):
        # Set up the environment variable for the test
        with patch.dict(os.environ, {"TEST_KEY": "test_value"}):
            result = self.provider.get("TEST_KEY")
            self.assertEqual(result, "test_value")

    def test_get_non_existing_env_var(self):
        # Ensure the environment variable does not exist
        with patch.dict(os.environ, {}, clear=True):
            result = self.provider.get("NON_EXISTENT_KEY")
            self.assertIsNone(result)

    def test_get_env_var_with_empty_value(self):
        # Set up the environment variable with an empty value
        with patch.dict(os.environ, {"EMPTY_KEY": ""}):
            result = self.provider.get("EMPTY_KEY")
            self.assertEqual(result, "")



import typing

from .exceptions import FieldConflictError, FieldValueNotProvidedError
from .field import AbstractField, Field, FieldPopulationResult
from .provider import AbstractProvider, EnvProvider


def _read_raw_value_from_file(path: str) -> str:
    with open(path, "r") as file:
        return file.read().strip()


class Minicfg:
    _prefix_attr_name = "_prefix"
    _default_provider_attr_name = "_default_provider"

    _prefix: str | None
    _default_provider: AbstractProvider | None

    _is_populated: bool

    def __init__(self, prefix: str | None = None, provider: AbstractProvider | None = None):
        # setup field names prefix:
        if prefix:
            self._prefix = prefix
        elif not hasattr(self, self._prefix_attr_name):
            # if prefix was not set by class decorator earlier:
            self._prefix = None

        # setup default provider:
        if provider:
            self._default_provider = provider
        elif not hasattr(self, self._default_provider_attr_name):
            # if default provider was not set by class decorator earlier:
            self._default_provider = None

        self._is_populated = False

    def __getattribute__(self, item: typing.Any):
        cls = super().__getattribute__("__class__")
        attr = getattr(cls, item, None)

        if isinstance(attr, AbstractField):
            return attr.value

        # default behavior for attributes that are not Fields
        return super().__getattribute__(item)

    @property
    def is_populated(self) -> bool:
        return self._is_populated

    def dict(self) -> dict[str, typing.Any]:
        if not self._is_populated:
            raise RuntimeError("config is not populated yet")

        result = {}
        for attr_name, attr in self._iter_public_attrs():
            if isinstance(attr, AbstractField):
                result[attr_name] = attr.value

            elif isinstance(attr, Minicfg):
                result[attr_name] = attr.dict()

        return result

    def populate(self) -> None:
        provider = self._default_provider or EnvProvider()

        for attr_name, attr in self._iter_public_attrs():
            if isinstance(attr, type) and issubclass(attr, Minicfg):  # if attr is a child Minicfg type
                child_minicfg = attr()

                # inherit prefix if needed:
                if self._prefix:
                    if child_minicfg._prefix:
                        child_minicfg._prefix = f"{self._prefix}{child_minicfg._prefix}"
                    else:
                        child_minicfg._prefix = self._prefix

                # inherit default provider if needed:
                if not child_minicfg._default_provider:
                    child_minicfg._default_provider = provider

                self.__setattr__(attr_name, child_minicfg)
                child_minicfg.populate()

            elif isinstance(attr, Field):  # if attr is field
                field = attr
                if not field.name:
                    field.name = attr_name

                file_field: Field | None = None
                file_field_result: FieldPopulationResult | None = None
                if field.attach_file_field:
                    file_field = Field(f"{field.name}_FILE")
                    file_field_result = file_field.populate_using_provider(
                        provider, file_field.name_with_prefix(self._prefix)
                    )

                field_result = field.populate_using_provider(provider, field.name_with_prefix(self._prefix))

                if field_result.populated and not field_result.default_value_used:
                    if file_field_result and file_field_result.populated:
                        raise FieldConflictError(
                            field_name=field.name_with_prefix(self._prefix),
                            file_field_name=file_field.name_with_prefix(self._prefix),
                            provider=provider,
                        )

                if file_field:  # if file field is attached to the original field
                    if file_field_result.populated:
                        file_field_raw_value = _read_raw_value_from_file(file_field.value)
                        field.populate_using_raw_value(file_field_raw_value)
                    else:
                        if not field_result.populated:
                            raise FieldValueNotProvidedError(field.name_with_prefix(self._prefix), provider)
                else:
                    if not field_result.populated:
                        raise FieldValueNotProvidedError(field.name_with_prefix(self._prefix), provider)

        self._is_populated = True

    def _iter_public_attrs(self):
        for attr_name in dir(self):
            if attr_name.startswith("_"):
                continue

            attr = super().__getattribute__(attr_name)
            yield attr_name, attr


def minicfg_provider(provider: AbstractProvider):
    def decorator(cls: Minicfg):
        cls._default_provider = provider
        return cls

    return decorator


def minicfg_prefix(prefix: str):
    def decorator(cls: Minicfg):
        cls._prefix = f"{prefix}_"
        return cls

    return decorator



import os
from abc import ABC, abstractmethod


class AbstractProvider(ABC):
    @abstractmethod
    def get(self, key: str) -> str | None:
        pass


class EnvProvider(AbstractProvider):
    def get(self, key: str) -> str | None:
        return os.getenv(key)



import unittest
from typing import Any

from minicfg.field import _NOT_SET, AbstractCaster, AbstractProvider, CastingError, Field, FieldPopulationResult


class MockCaster(AbstractCaster):
    def cast(self, value: str) -> Any:
        if value == "invalid":
            raise ValueError("Invalid value")
        return value.upper()


class MockProvider(AbstractProvider):
    def __init__(self, values):
        self._values = values

    def get(self, key: str) -> str | None:
        return self._values.get(key)


class TestField(unittest.TestCase):
    def test_initialization_and_properties(self):
        field = Field(name="test_field", default="default_value", caster=MockCaster(), attach_file_field=True)

        self.assertEqual(field.name, "test_field")
        self.assertEqual(field._default, "default_value")
        self.assertIsInstance(field._caster, MockCaster)
        self.assertTrue(field.attach_file_field)
        self.assertIs(field.value, _NOT_SET)

    def test_name_with_prefix(self):
        field = Field(name="field_name")
        self.assertEqual(field.name_with_prefix("prefix_"), "prefix_field_name")
        self.assertEqual(field.name_with_prefix(None), "field_name")

    def test_populate_using_raw_value(self):
        field = Field(caster=MockCaster())

        result = field.populate_using_raw_value("raw_value")
        self.assertEqual(field.value, "RAW_VALUE")
        self.assertTrue(result.populated)

        field = Field(caster=MockCaster())
        with self.assertRaises(CastingError):
            field.populate_using_raw_value("invalid")

    def test_populate_using_provider(self):
        provider = MockProvider(values={"field_name": "provider_value"})
        field = Field(name="field_name", default="default_value", caster=MockCaster())

        result = field.populate_using_provider(provider, "field_name")
        self.assertEqual(field.value, "PROVIDER_VALUE")
        self.assertTrue(result.populated)

        provider = MockProvider(values={})
        field = Field(name="field_name", default="default_value", caster=MockCaster())

        result = field.populate_using_provider(provider, "field_name")
        self.assertEqual(field.value, "default_value")
        self.assertTrue(result.populated)
        self.assertTrue(result.default_value_used)

        result = field.populate_using_provider(provider, "non_existent_field")
        self.assertTrue(result.populated)
        self.assertTrue(result.default_value_used)


if __name__ == "__main__":
    unittest.main()



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any

from .caster import AbstractCaster
from .exceptions import CastingError
from .provider import AbstractProvider


@dataclass
class FieldPopulationResult:
    populated: bool
    default_value_used: bool = False


class AbstractField(ABC):
    @property
    @abstractmethod
    def name(self) -> str | None:
        pass

    @name.setter
    @abstractmethod
    def name(self, value: str) -> None:
        pass

    @property
    @abstractmethod
    def value(self) -> Any:
        pass

    @value.setter
    @abstractmethod
    def value(self, value: Any) -> None:
        pass

    @property
    @abstractmethod
    def attach_file_field(self) -> bool:
        pass

    @abstractmethod
    def name_with_prefix(self, prefix: str | None) -> str:
        pass

    @abstractmethod
    def populate_using_raw_value(self, raw_value: str) -> FieldPopulationResult:
        pass

    @abstractmethod
    def populate_using_provider(self, provider: AbstractProvider, field_name: str) -> FieldPopulationResult:
        pass


_NOT_SET = object()


class Field(AbstractField):
    _name: str
    _name_prefix: str | None

    _default: Any
    _caster: AbstractCaster
    _attach_file_field: bool

    _value: Any

    def __init__(
        self,
        name: str | None = None,
        default: Any = _NOT_SET,
        caster: AbstractCaster | None = None,
        attach_file_field: bool = False,
    ):
        self._name = name

        self._default = default
        self._caster = caster
        self._attach_file_field = attach_file_field

        self._value = _NOT_SET

    @property
    def name(self) -> str | None:
        return self._name

    @name.setter
    def name(self, value: str) -> None:
        self._name = value

    @property
    def value(self) -> Any:
        return self._value

    @value.setter
    def value(self, value: Any) -> None:
        self._value = value

    @property
    def attach_file_field(self) -> bool:
        return self._attach_file_field

    def name_with_prefix(self, prefix: str | None) -> str:
        if not prefix:
            return self._name

        return f"{prefix}{self._name}"

    def populate_using_raw_value(self, raw_value: str) -> FieldPopulationResult:
        if self._caster:
            try:
                self._value = self._caster.cast(raw_value)
                return FieldPopulationResult(populated=True)
            except Exception as e:
                raise CastingError(field_name=self.name, raw_value=raw_value, caster=self._caster, exception=e)

        self._value = raw_value
        return FieldPopulationResult(populated=True)

    def populate_using_provider(self, provider: AbstractProvider, field_name: str) -> FieldPopulationResult:
        raw_value = provider.get(field_name)
        if raw_value is None:
            if self._default is not _NOT_SET:
                self._value = self._default
                return FieldPopulationResult(populated=True, default_value_used=True)
            return FieldPopulationResult(populated=False)

        return self.populate_using_raw_value(raw_value)
